#!/bin/bash
########################################################################
# build_hdf5.sh
#
# Build hdf5 and package for UPS.
#
# Adapted from build.sh.template provided by ssibuildshims v2_04_01.
########################################################################

########################################################################
# Non-customize-able boilerplate: skip to "Customize-able functions."
########################################################################

####################################
# Useful variables.
prog=${BASH_SOURCE[0]##*/}

##################
# Local (dumb) ssi_die() function for use until we source ssi_functions.
ssi_die() {
  printf "FATAL ERROR: ${*}\n" 1>&2; exit 1
}
##################

##################
# Basic usage function for use by ensure_buildshims until we have enough
# information to provide something more specific. Do NOT customize!
build_script_usage() {
    cat <<EOF
USAGE: ${prog} <ssi-msg-config-options> [--] <product-dir> <qualifier-args> <maketar>
       ${prog} -h|--help [<product-dir>]

  Build ${package} and package for UPS.

OPTIONS

  -h|--help

    Help. Supply <product-dir> for specific, detailed information on
    invoking ${prog}.

  <ssi-msg-config-options>

    Supply <product-dir> for more information.

ARGUMENTS

  <product-dir>

    The top-level UPS directory in which the source code distribution
    was unpacked. When specified with -h|--help, this enables us to
    provide more specific help about invoking ${prog}.

  <qualifier-args>
  <maketar>

    Supply <product-dir> for more information.

EOF
}
##################

##################
# Make sure we have UPS and ssibuildshims set up.
ensure_ssibuildshims() {
  if ! (( $# )) || [ "${@: -1}" = "-h" ] || [ "${@: -1}" = "--help" ]; then
    build_script_usage
    exit 1
  fi
  # Find <product-dir> from our arguments:
  local arg
  for arg in "${@}"; do
    [[ "$arg" ==  -* ]] && continue;
    product_dir="${arg}"; break
  done
  if [ -z "${product_dir}" ]; then
    # Bad usage.
    ssi_die "Bad <product-dir> ${product_dir}\n$(build_script_usage)"
    exit 2
  elif ! { [ -d "${product_dir}/.upsfiles" ] && [ -w "${product_dir}" ]; }; then
    # Something wrong with <product-dir>.
    ssi_die "<product-dir> not a valid, writable unified UPS directory: " \
      "${product_dir}\n$(build_script_usage)"
  fi
  # Promising...
  local abs_product_dir="$(cd "${product_dir}" 2>/dev/null 2>&1 && pwd -P)"
  (( $? )) && ssi_die "unable to cd to <product-dir>: ${product_dir}\n" \
    "$(build_script_usage)"
  # Find our own bootstraps and yank 'em.
  local ssibuildshims_topdir="${product_dir}/ssibuildshims/${ssibuildshims_version}"
  [ -f "${ssibuildshims_topdir}/bin/ssi_functions" ] && \
    source "${ssibuildshims_topdir}/bin/ssi_functions" >/dev/null 2>&1 && \
    ensure_ups "${product_dir}" && \
    setup ssibuildshims ${ssibuildshims_version} -z "${product_dir}" || \
    ssi_die "unable to find ssibuildshims ${ssibuildshims_version} in
${product_dir} and set it up with UPS."
  # Absolute is safer.
  absolute_path -- "${product_dir}" product_dir
}
##################

########################################################################
# Customize-able functions.
########################################################################

##################
# Ascertain whether we built and installed correctly. Add more /
# different calls to verify_in_filesystem as appropriate. Arguments
# (e.g. -q) are passed through to verify_in_filesystem.
is_built() { # *C*
  ${trace} verify_in_filesystem -d "${@}" "${pkgdir}/lib"
}
##################

##################
# If a distribution archive is wanted and the installation is successful
# according to is_built(), then make it. Any arguments (e.g. -q) are
# passed through both to is_built() and make_basic_distribution().
#
# If make_basic_distribution() finds that the product is declared
# correctly and attempts to make the distribution, then it will cause us
# to exit with the status code of that operation. If not, it returns an
# error code.
#
# See make_basic_distribution -h for useful options to add to
# mbd_opts(). *C*
maybe_make_distribution() {
  if [ "${maketar}" = "tar" ]; then
    # Customize mbd_opts. *C*
    local mbd_opts=()
    ${trace} is_built "$@" && \
      ${trace} make_basic_distribution "${@}" "${mbd_opts[@]}"
  fi
}
##################

########################################################################
# Package-specific functions.
########################################################################

# ...

########################################################################
# Main.
########################################################################

####################################
# Useful variables. *C*
trace=maybe_xtrace # Could set to xtrace instead for unconditional.
(( ssi_trace = 1 )) # Trace certain commands by default.
####################################

####################################
# Package name and version information. *C*
package=hdf5
origpkgver=v1_12_0
pkgver=${origpkgver}a # Append letters for local patches.
ssibuildshims_version=v2_07_00
pkgdotver=${origpkgver//_/.}; pkgdotver=${pkgdotver#v}
####################################

##################
# Other useful variables. *C*
pkgtarname=${package}-${pkgdotver}
pkgtarfile=${pkgtarname}.tar.bz2
##################

####################################
# Handle options and arguments.

# Set up ssibuildshims and UPS / product_dir as early as possible.
ensure_ssibuildshims "${@}"

##################
# Now parse options and arguments, setting basequal, extraqual, maketar,
# etc.
#
# Use verify_quals() to parse command line options, verify that the user
# has provided qualifiers supported by this build script, and set
# expected variables such as basequal, extraqual, cqual, etc.
#
# Customize the extra usage text as appropriate. Note that DEBUG_PATCHES
# and VERBOSE_PATCHES are honored by prepare_source() and
# patch_source(). DEBUG_(CONFIG|BUILD|TEST|INSTALL) should be accounted
# for (or not) by the relevant command in this script.
#
# See verify_quals -h for options and examples.
verify_quals \
  --cqual={e17,e19,e20,c2,c7,} \
  --bq-mqual={,nofcpp,mpich,openmpi,mvapich} \
  --btype={debug,prof,} \
  --parse --usage-file=- -- "$@" <<EOF || exit # *C*
ARGUMENTS

  N.B. If nofcpp is present, then build-type and cqual must be absent;
       otherwise they must be present.

ENVIRONMENT

  DEBUG_(CONFIG|BUILD)

    Verbose output from the corresponding step.

  DEBUG_PATCHES
  VERBOSE_PATCHES

    Command tracing and increased verbosity for source unpacking and patching
    operations. DEBUG_PATCHES will cause exit after patches have been applied.

  WANT_TESTS

    Set to 1 to enable tests (unset => skip tests).
EOF
##################

##################
# Define globals: pkgdir, sourcedir, blddir, fullqual, etc. See
# define_basics -h for steering options.

# Check for invalid qualifier combinations.
if [ "$mqual" = "nofcpp" ]; then
  [ -z "$cqual$btype" ] || \
    ssi_die "If nofcpp is specified, then cqual ($cqual) and btype ($btype) may not be"
elif [ -z "$cqual$btype" ]; then
  ssi_die "cqual and btype must be specified if nofcpp is not"
fi

db_args=(TRACK_PROGRESS)

[ -n "$btype" ] || db_args+=(NO_TYPE)
[ "$mqual" = nofcpp ] && db_args+=(NO_DEBUG_SOURCE)

${trace} define_basics "${db_args[@]}" || exit # *C*
##################

####################################
# Declare ourselves privately, and then set ourselves up to allow setup
# of dependencies. *C*
#
# Comment or delete if not required:
${trace} fake_declare_product "${ssi_args[@]}" -x -- || exit
####################################

####################################
# Build-only dependencies. *C*
#
${trace} setup_cmake --latest || exit
${trace} setup_ninja --optional && (( want_ninja = 1 ))
####################################

# Print active products.
echo
ups active
echo

# Possible short-circuit if we're already built and declared.
${trace} maybe_make_distribution --silent

ssi_info "building ${package} for ${OS}-${plat}-${qualdir} (flavor ${pkgflvr})"

# Ensure directories: pkgdir, sourcedir, blddir.
${trace} init_pkg_dirs || exit

####################################
# Prepare source for building. *C*

${trace} cd "${sourcedir}" \
  || ssi_die "unable to cd to source directory ${sourcedir}"

# Source tarballs and patches. See prepare_source -h / patch_source -h
# for options.
${trace} prepare_source "${ssi_args[@]}" -P "${patchdir}" \
  "${tardir}/${pkgtarfile}" \
  || exit

# Bail early if DEBUG_PATCHES is set.
[ -n "${DEBUG_PATCHES}" ] && exit 1
####################################

####################################
# Prepare to build. *C*
#
# Set configure / compile flags, environment variables, etc.

# Later options (e.g. compiler-specific ones) can override earlier ones.
declare -a cmake_options cflg fflg cxxflg

# Determine and save the CMake build type, since we need it in a couple
# of places.
case $btype in
  debug) cmake_build_type=Debug
    cmake_options+=( \
      -DHDF5_ENABLE_USING_MEMCHECKER:BOOL=ON \
      -DHDF5_MEMORY_ALLOC_SANITY_CHECK:BOOL=ON \
      -DHDF5_ENABLE_DEBUG_APIS:BOOL=ON \
      -DHDF5_ENABLE_TRACE:BOOL=ON \
      );;
  prof) cmake_build_type=RelWithDebInfo;;
  "") cmake_build_type=Release;;
  *) ssi_die -I "accepted unknown btype=$btype"
esac

case $cqual in
  "") cstd=99;;
  c[27]) cstd=11; cxxstd=17; cc=clang; fc=gfortran; cxx=clang++;;
  e1[79]|e20) cstd=11; cxxstd=17; cc=gcc; fc=gfortran; cxx=g++;;
  e21) cstd=11; cxxstd=20; cc=gcc; fc=gfortran; cxx=g++;;
  *) ssi_die -I "accepted unknown cqual=$cqual"
esac

fstd=legacy
fflg+=(-std=$fstd)

##################
# Main CMake options first.
(( want_ninja )) && cmake_options+=(-GNinja)

cmake_options=( \
  -DCMAKE_INSTALL_PREFIX:PATH=${pkgdir} \
  ${cmake_build_type:+-DCMAKE_BUILD_TYPE:STRING=${cmake_build_type}} \
  "${cmake_options[@]}" \
  )
##################

##################
# HDF5 options and features.

# Options and feature settings common to all configurations.
cmake_options+=( \
  -DBUILD_STATIC_LIBS:BOOL=OFF \
  -DBUILD_TESTING:BOOL=$(( WANT_TESTS )) \
  -DHDF5_ENABLE_ALL_WARNINGS:BOOL=ON \
  -DHDF5_BUILD_CPP_LIB:BOOL=OFF \
  -DHDF5_BUILD_EXAMPLES:BOOL=ON \
  -DHDF5_BUILD_JAVA:BOOL=OFF \
  -DHDF5_ENABLE_Z_LIB_SUPPORT:BOOL=ON \
  -DHDF5_NO_PACKAGES:BOOL=OFF \
  -DHDF5_PACKAGE_EXTLIBS:BOOL=OFF \
  -DONLY_SHARED_LIBS:BOOL=ON \
)

case $mqual in
  nofcpp)
    cmake_options+=( \
      -DHDF5_BUILD_FORTRAN:BOOL=OFF \
      -DHDF5_BUILD_HL_LIB:BOOL=OFF \
      -DHDF5_ENABLE_PARALLEL:BOOL=OFF \
      -DHDF5_ENABLE_THREADSAFE:BOOL=ON \
      );;
  mpich|mvapich2|openmpi)
    cmake_options+=( \
      -DHDF5_BUILD_FORTRAN:BOOL=ON \
      -DHDF5_BUILD_HL_LIB:BOOL=ON \
      -DHDF5_ENABLE_PARALLEL:BOOL=ON \
      -DHDF5_ENABLE_THREADSAFE:BOOL=OFF \
      )
    [ -n "$fc" ] && fc=mpif90
    [ -n "$cc" ] && cc=mpicc
    [ -n "$cxx" ] && cxx=mpicxx;;
  "") cmake_options+=( \
    -DHDF5_BUILD_FORTRAN:BOOL=OFF \
    -DHDF5_BUILD_HL_LIB:BOOL=ON \
    -DHDF5_ENABLE_PARALLEL:BOOL=OFF \
    -DHDF5_ENABLE_THREADSAFE:BOOL=OFF \
    );;
  *) ssi_die -I "accepted unknown mqual=$mqual"
esac

##################
# Compiler options.
cmake_options+=( \
  ${fc:+-DCMAKE_Fortran_COMPILER=${fc}} \
  ${cc:+-DCMAKE_C_COMPILER=${cc}} \
  ${cstd:+-DCMAKE_C_STANDARD=${cstd} \
  -DCMAKE_C_STANDARD_REQUIRED:BOOL=ON \
  -DCMAKE_C_EXTENSIONS:BOOL=OFF} \
  ${cxx:+-DCMAKE_CXX_COMPILER=${cxx}} \
  ${cxxstd:+-DCMAKE_CXX_STANDARD=${cxxstd} \
  -DCMAKE_CXX_STANDARD_REQUIRED:BOOL=ON \
  -DCMAKE_CXX_EXTENSIONS:BOOL=OFF} \
  ${fflg[*]:+-DCMAKE_Fortran_FLAGS="${fflg[*]}"} \
  ${cflg[*]:+-DCMAKE_C_FLAGS="${cflg[*]}"} \
  ${cxxflg[*]:+-DCMAKE_CXX_FLAGS="${cxxflg[*]}"} \
  )

##################
# Configure parallelism. *C*
ncores=$(ncores)

# For cmake --build:
[[ -n "${CMAKE_BUILD_PARALLEL_LEVEL}" ]] || \
  export CMAKE_BUILD_PARALLEL_LEVEL=${ncores}

# For ctest:
[[ -n "${CTEST_PARALLEL_LEVEL}" ]] || \
  export CTEST_PARALLEL_LEVEL=${ncores}
##################
####################################

####################################
# Configure, build, and install.
#
# Use ${blddir}/${pkgtarname} for in-source builds, ${blddir} otherwise.

${trace} cd "${blddir}" || \
  ssi_die "unable to cd to build directory ${blddir}"

##################
# Configure. *C*
${trace} cmake "${cmake_options[@]}" \
  ${DEBUG_CONFIG:+--verbose} \
  "${sourcedir}/${pkgtarname}" || \
  ssi_die "configure failed"
##################

##################
# Build. *C*
${trace} cmake --build "${blddir}" ${DEBUG_BUILD:+--verbose} || \
  ssi_die "build failed"
##################

##################
# Test. *C*
(( WANT_TESTS )) && \
  { ${trace} ctest --output-on-failure || ssi_die "tests failed"; }
##################

##################
# Install. *C*

${trace} cmake --install "${blddir}" || ssi_die "install failed"
##################

ssi_info "finished building ${package} ${pkgver}"

####################################
# Pre-declaration processing. *C*
#
# Move things, install from contrib, clean build products from in-source
# builds, etc.
####################################

# 1.10.1 installed unwanted libraries.
rm -f "${pkgdir}/lib/libdynlib"*

# Fix rpaths in Fortran libraries.
if [[ "${OS1}" == "Darwin" ]]; then
  flibs=(libhdf5_{,hl_}fortran.dylib)
  pushd "${pkgdir}/lib"
  for flib in "${flibs[@]}"; do
    if [[ -f "${pkgdir}/lib/${flib}" ]]; then
      otool -L "${flib}"
      fix_macos_lib_symbols "${flib}"
    fi
  done
  popd
fi

ssi_info "post-processing for relocatability"
${trace} fix_pc && ${trace} smoke_test_pc && \
  ${trace} fix_config_script -i -- \
  =GCC_FQ_DIR =CLANG_FQ_DIR -- \
  "${pkgdir}"/bin/h5{{,hl,p}cc,c++,fc} && \
  ${trace} rm_libtool_la || ssi_die "post-processing failed"

##################
# Declare and set up the installed package. *C*
#
# Add -c for current declaration if required.
${trace} declare_product -x -- || exit
##################

####################################
# Post-declaration processing. *C*
#
# Operations requiring the *final* declaration and setup (as opposed to
# the pre-build fake declaration and setup). This should be rare.
####################################

#################
# Verify installation and make distribution archive. *C*
#
# To customize verification, see is_built(), since it is also used by
# maybe_make_distribution(). See the documentation for those functions
# (above) for customization opportunities.

# Add a message to ssi_build_failed if appropriate.
${trace} is_built && ${trace} ssi_build_completed && \
  ssi_info "${package} is installed at ${pkgdir}" && \
  ${trace} maybe_make_distribution || \
  ssi_build_failed # *C*
########################################################################
# Nothing beyond this point.
########################################################################
